type WIND
  ,128
XLISP version 2.0
41 4 339 508
InVisible GoAway
0
0

TYPE WIND
  ,129
Graphics Window
22 4 254 508
InVisible NoGoAway
2
0

TYPE DLOG
  ,129
About XLISP
50 100 290 395
Visible NoGoAway
3
0
129

TYPE DITL
  ,129
9

staticText
20 20 40 275
XLISP v2.0, February 6, 1988

staticText
40 20 60 275
Copyright (c) 1988, by David Betz

staticText
60 20 80 275
All Rights Reserved

staticText
90 20 110 275
Author contact information:

staticText
110 40 130 275
David Betz

staticText
130 40 150 275
127 Taylor Road

staticText
150 40 170 275
Peterborough, NH  03458

staticText
170 40 190 275
(603) 924-6936

staticText
200 20 220 275
Portions Copyright Think Technologies

TYPE MENU
  ,1
\14
About XLISP
(-

TYPE MENU
  ,256
File
Load.../L
Load Noisily.../N
(-
Quit/Q

TYPE MENU
  ,257
Edit
Undo/Z
(-
Cut/X
Copy/C
Paste/V
Clear

TYPE MENU
  ,258
Control
Break/B
Continue/P
Clean Up Error/G
Cancel Input/U
Top Level/T
(-
Split Screen/S

