/* samples.h -- convert sound (prefix) to lisp array */

sound_type snd_from_array(double t0, double sr, LVAL array);
  /* LISP: (SND-FROM-ARRAY ANYNUM ANYNUM ANY) */

LVAL snd_samples(sound_type s, long len); /* LISP: (SND-SAMPLES SOUND FIXNUM) */
long snd_length(sound_type s, long len); /* LISP: (SND-LENGTH SOUND FIXNUM) */

double snd_maxsamp(sound_type s); /* LISP: (SND-MAXSAMP SOUND) */
