(setf transfiles '("amosc" "const" "fmosc" "integrate" "log" "osc" "prod"
		   "pwl" "recip" "upsample" "scale" "sine" "partial" "white"
		   "tone" "tonev" "atonev" "atone" "reson" "areson"
		   "resonvc" "resoncv" "aresonvc" "aresoncv" "resonvv"
		   "aresonvv" "offset" "slope" "delay" "delaycv" "shape"
                   "sampler" "exp"))

; deleted comb
