sound_type snd_make_white(time_type t0, rate_type sr, time_type d);
sound_type snd_white(time_type t0, rate_type sr, time_type d);
    /* LISP: (snd-white ANYNUM ANYNUM ANYNUM) */
#include <math.h>

/* random returns from 0 to 0x7FFFFFF.  Scale and offset
 * to get range from -1 to +1
 */
#define random_scale (2.0/0x7FFFFFFF)
