/*
 * stats.c
 *
 * produce statistics.
 */

#include <stdio.h>
#ifndef mips
#include "stdlib.h"
#endif
#ifdef XLISP_SOUND
#include "xlisp.h"
#endif
#include "sound.h"
#include "falloc.h"


void stats()
{
    fprintf(stdout, "\n\nNyquist statistics:\n\n");
    fprintf(stdout, "Memory usage:\n");
    fprintf(stdout, "\tconsumed %d pools of size %d\n", npools, MAXPOOLSIZE);
    fprintf(stdout, "\tdata structure usage:\n");
    fprintf(stdout, "\t\tsounds\t%d\n", sound_used);
    fprintf(stdout, "\t\tsnd lists\t%d\n", snd_list_used);
    fprintf(stdout, "\t\tsample blocks\t%d\n", sample_block_used);
    fprintf(stdout, "\n");
}
