resource 'MENU' (1) {
	1,
	textMenuProc,
	0x7FFFFFFD,
	enabled,
	apple,
	{	/* array: 2 elements */
		/* [1] */
		"About XLISP", noIcon, noKey, noMark, plain,
		/* [2] */
		"-", noIcon, noKey, noMark, plain
	}
};

resource 'MENU' (256) {
	256,
	textMenuProc,
	0x7FFFFFFB,
	enabled,
	"File",
	{	/* array: 4 elements */
		/* [1] */
		"Load...", noIcon, "L", noMark, plain,
		/* [2] */
		"Load Noisily...", noIcon, "N", noMark, plain,
		/* [3] */
		"-", noIcon, noKey, noMark, plain,
		/* [4] */
		"Quit", noIcon, "Q", noMark, plain
	}
};

resource 'MENU' (257) {
	257,
	textMenuProc,
	0x7FFFFFFD,
	enabled,
	"Edit",
	{	/* array: 6 elements */
		/* [1] */
		"Undo", noIcon, "Z", noMark, plain,
		/* [2] */
		"-", noIcon, noKey, noMark, plain,
		/* [3] */
		"Cut", noIcon, "X", noMark, plain,
		/* [4] */
		"Copy", noIcon, "C", noMark, plain,
		/* [5] */
		"Paste", noIcon, "V", noMark, plain,
		/* [6] */
		"Clear", noIcon, noKey, noMark, plain
	}
};

resource 'MENU' (258) {
	258,
	textMenuProc,
	0x7FFFFFDF,
	enabled,
	"Control",
	{	/* array: 7 elements */
		/* [1] */
		"Break", noIcon, "B", noMark, plain,
		/* [2] */
		"Continue", noIcon, "P", noMark, plain,
		/* [3] */
		"Clean Up Error", noIcon, "G", noMark, plain,
		/* [4] */
		"Cancel Input", noIcon, "U", noMark, plain,
		/* [5] */
		"Top Level", noIcon, "T", noMark, plain,
		/* [6] */
		"-", noIcon, noKey, noMark, plain,
		/* [7] */
		"Split Screen", noIcon, "S", noMark, plain
	}
};

resource 'MENU' (259) {
	259,
	textMenuProc,
	0x7FFFFB7B,
	enabled,
	"Editor",
	{	/* array: 12 elements */
		/* [1] */
		"Editor to Top",  noIcon, noKey, noMark, plain,
		/* [2] */
		"-", noIcon, noKey, noMark, plain,
		/* [3] */
		"New", noIcon, "N", noMark, plain,
		/* [4] */
		"Open", noIcon, "O", noMark, plain,
		/* [5] */
		"Close", noIcon, "W", noMark, plain,
		/* [6] */
		"-", noIcon, noKey, noMark, plain,
		/* [7] */
		"Save", noIcon, "S", noMark, plain,
		/* [8] */
		"Save AsÉ", noIcon, noKey, noMark, plain,
		/* [9] */
		"Revert", noIcon, noKey, noMark, plain,
		/* [10] */
		"-", noIcon, noKey, noMark, plain,
		/* [11] */
		"Page SetupÉ", noIcon, noKey, noMark, plain,
		/* [12] */
		"PrintÉ", noIcon, "P", noMark, plain
	}
};

resource 'DITL' (129) {
	{	/* array DITLarray: 9 elements */
		/* [1] */
		{20, 20, 40, 275},
		StaticText {
			enabled,
			"XLISP v2.0, February 6, 1988"
		},
		/* [2] */
		{40, 20, 60, 275},
		StaticText {
			enabled,
			"Copyright (c) 1988, by David Betz"
		},
		/* [3] */
		{60, 20, 80, 275},
		StaticText {
			enabled,
			"All Rights Reserved"
		},
		/* [4] */
		{90, 20, 110, 275},
		StaticText {
			enabled,
			"Author contact information:"
		},
		/* [5] */
		{110, 40, 130, 275},
		StaticText {
			enabled,
			"David Betz"
		},
		/* [6] */
		{130, 40, 150, 275},
		StaticText {
			enabled,
			"127 Taylor Road"
		},
		/* [7] */
		{150, 40, 170, 275},
		StaticText {
			enabled,
			"Peterborough, NH  03458"
		},
		/* [8] */
		{170, 40, 190, 275},
		StaticText {
			enabled,
			"(603) 924-6936"
		},
		/* [9] */
		{200, 20, 220, 275},
		StaticText {
			enabled,
			"Portions Copyright Think Technologies"
		}
	}
};

resource 'DLOG' (129) {
	{50, 100, 290, 395},
	altDBoxProc,
	-1,
	noGoAway,
	0x0,
	129,
	"About XLISP"
};

resource 'WIND' (128) {
	{41, 4, 339, 508},
	documentProc,
	invisible,
	-1,
	0x0,
	"XLISP version 2.0"
};

resource 'WIND' (129) {
	{22, 4, 254, 508},
	plainDBox,
	invisible,
	noGoAway,
	0x0,
	"Graphics Window"
};

