 (SETF MAX-STOP-TIME 10E20) 

 (SETF MIN-START-TIME -10E20) 

 (setf snd-head-none 0) 

 (setf snd-head-AIFF 1) 

 (setf snd-head-IRCAM 2) 

 (setf snd-head-NeXT 3) 

 (setf snd-head-Wave 4) 

 (setf snd-head-channels 1) 

 (setf snd-head-mode 2) 

 (setf snd-head-bits 4) 

 (setf snd-head-srate 8) 

 (setf snd-head-dur 16) 

 (setf snd-head-mode-adpcm 0) 

 (setf snd-head-mode-pcm 1) 

 (setf snd-head-mode-ulaw 2) 

 (setf snd-head-mode-alaw 3) 

 (setf snd-head-mode-float 4) 

 (setf snd-head-mode-upcm 5) 

