sound_type snd_make_amosc(sound_type s, double step, rate_type sr, double hz, time_type t0, sound_type am, double phase);
sound_type snd_amosc(sound_type s, double step, rate_type sr, double hz, time_type t0, sound_type am, double phase);
    /* LISP: (snd-amosc SOUND ANYNUM ANYNUM ANYNUM ANYNUM SOUND ANYNUM) */

#define table_type sound_type
